@@ -279,6 +279,18 @@ class ModelInference:
         control_population_spatial_posterior_beta = (neighborhood_imputation_prior_strength*negative_control_guide_control_population_total_normalized_counts_reps[replicate_indices]) + (neighborhood_imputation_likelihood_strength * each_guide_control_population_spatial_contribution_reps)
         imputation_posterior_beta = control_population_spatial_posterior_beta
 
+        max_imputation_posterior = 1000 # Setting this as max value, since calculating the KL divergence of a very high posterior value may result in a precision error
+        for rep_i in range(len(replicate_indices)):
+            if max(imputation_posterior_alpha[rep_i], imputation_posterior_beta[rep_i]) > max_imputation_posterior:
+                print("Downscaling posterior: {}, {}".format(imputation_posterior_alpha[rep_i],imputation_posterior_beta[rep_i]))
+                print("Other params: Prior strength={}, likelihood_strength={}".format(neighborhood_imputation_prior_strength,neighborhood_imputation_likelihood_strength ))
+                if imputation_posterior_alpha[rep_i] > imputation_posterior_beta[rep_i]:
+                    downscale_factor = max_imputation_posterior/imputation_posterior_alpha[rep_i]
+                else:
+                    downscale_factor = max_imputation_posterior/imputation_posterior_beta[rep_i]
+                imputation_posterior_alpha[rep_i] = imputation_posterior_alpha[rep_i] * downscale_factor
+                imputation_posterior_beta[rep_i] = imputation_posterior_beta[rep_i] * downscale_factor
+
         return imputation_posterior_alpha, imputation_posterior_beta, each_guide_sample_population_spatial_contribution_reps, each_guide_control_population_spatial_contribution_reps
 
     @staticmethod
@@ -325,59 +337,65 @@ class ModelInference:
         return shrinkage_result
 
 
+
+
+
 #-------------------------Optimization functions
-class SingletonImputationOptimizer:
 
-    @staticmethod
-    def objective_function_of_each_guide_set(rep_i, 
-                                            guide_set: List[Guide], 
-                                            negative_control_guide_sample_population_total_normalized_counts_reps: List[float], 
-                                            negative_control_guide_control_population_total_normalized_counts_reps: List[float],
-                                            params: float):
-        singleton_imputation_prior_strength_test = params
-        KL_guide_imputation_score_total: float = 0
-
-        for each_guide in guide_set:
-            # Get the posterior
-            imputation_posterior_alpha, imputation_posterior_beta = ModelInference.perform_singleton_score_imputation(each_guide, 
-                negative_control_guide_sample_population_total_normalized_counts_reps, 
-                negative_control_guide_control_population_total_normalized_counts_reps, 
-                singleton_imputation_prior_strength_test,
-                [rep_i])
-
-            imputation_posterior_alpha = imputation_posterior_alpha[0]
-            imputation_posterior_beta = imputation_posterior_beta[0]
-
-            true_alpha = each_guide.sample_population_normalized_count_reps[rep_i]
-            true_beta = each_guide.control_population_normalized_count_reps[rep_i]
-
-            # Calculate KL divergence between the posterior and the likelihood
-            KL_guide_imputation_score: float = StatisticalHelperMethods.KL_beta(true_alpha, true_beta, imputation_posterior_alpha, imputation_posterior_beta)
-            
-            # Add score to the main placeholder to get the final sum
-            KL_guide_imputation_score_total = KL_guide_imputation_score_total + KL_guide_imputation_score 
-            
-        return KL_guide_imputation_score_total, len(guide_set)
-
-    @staticmethod
-    def objective_function_of_all_guides(rep_i: int,
-                                        experiment_guide_sets: ExperimentGuideSets,
-                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+def singleton_imputation_objective_function_of_each_guide_set(rep_i, 
+                                        guide_set: List[Guide], 
+                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float], 
                                         negative_control_guide_control_population_total_normalized_counts_reps: List[float],
                                         params: float):
-        KL_guide_imputation_score_total_negative, negative_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.negative_control_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
-        KL_guide_imputation_score_total_positive, positive_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.positive_control_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
-        KL_guide_imputation_score_total_observation, observation_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.observation_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
+    singleton_imputation_prior_strength_test = params
+    KL_guide_imputation_score_total: float = 0
+
+    for each_guide in guide_set:
+        # Get the posterior
+        imputation_posterior_alpha, imputation_posterior_beta = ModelInference.perform_singleton_score_imputation(each_guide, 
+            negative_control_guide_sample_population_total_normalized_counts_reps, 
+            negative_control_guide_control_population_total_normalized_counts_reps, 
+            singleton_imputation_prior_strength_test,
+            [rep_i])
+
+        imputation_posterior_alpha = imputation_posterior_alpha[0]
+        imputation_posterior_beta = imputation_posterior_beta[0]
+
+        true_alpha = each_guide.sample_population_normalized_count_reps[rep_i]
+        true_beta = each_guide.control_population_normalized_count_reps[rep_i]
+
+        # Calculate KL divergence between the posterior and the likelihood
+        KL_guide_imputation_score: float = StatisticalHelperMethods.KL_beta(true_alpha, true_beta, imputation_posterior_alpha, imputation_posterior_beta)
         
-        total_n = (negative_total_n+positive_total_n+observation_total_n)
+        # Add score to the main placeholder to get the final sum
+        KL_guide_imputation_score_total = KL_guide_imputation_score_total + KL_guide_imputation_score 
         
-        
-        KL_guide_imputation_score_total_negative_avg = np.inf if negative_total_n == 0 else KL_guide_imputation_score_total_negative/negative_total_n
-        KL_guide_imputation_score_total_positive_avg = np.inf if positive_total_n == 0 else KL_guide_imputation_score_total_positive/positive_total_n
-        KL_guide_imputation_score_total_observation_avg = np.inf if observation_total_n == 0 else KL_guide_imputation_score_total_observation/observation_total_n
-        KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (KL_guide_imputation_score_total_negative + KL_guide_imputation_score_total_positive + KL_guide_imputation_score_total_observation) / total_n
-        
-        return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
+    return KL_guide_imputation_score_total, len(guide_set)
+
+def singleton_imputation_objective_function_of_all_guides(rep_i: int,
+                                    experiment_guide_sets: ExperimentGuideSets,
+                                    negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+                                    negative_control_guide_control_population_total_normalized_counts_reps: List[float],
+                                    params: float):
+    KL_guide_imputation_score_total_negative, negative_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.negative_control_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
+    KL_guide_imputation_score_total_positive, positive_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.positive_control_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
+    KL_guide_imputation_score_total_observation, observation_total_n = SingletonImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.observation_guides, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
+    
+    total_n = (negative_total_n+positive_total_n+observation_total_n)
+    
+    
+    KL_guide_imputation_score_total_negative_avg = np.inf if negative_total_n == 0 else KL_guide_imputation_score_total_negative/negative_total_n
+    KL_guide_imputation_score_total_positive_avg = np.inf if positive_total_n == 0 else KL_guide_imputation_score_total_positive/positive_total_n
+    KL_guide_imputation_score_total_observation_avg = np.inf if observation_total_n == 0 else KL_guide_imputation_score_total_observation/observation_total_n
+    KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (KL_guide_imputation_score_total_negative + KL_guide_imputation_score_total_positive + KL_guide_imputation_score_total_observation) / total_n
+    
+    return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
+
+
+
+class SingletonImputationOptimizer:
+    singleton_imputation_objective_function_of_each_guide_set = staticmethod(singleton_imputation_objective_function_of_each_guide_set)
+    singleton_imputation_objective_function_of_all_guides = staticmethod(singleton_imputation_objective_function_of_all_guides)
 
     @staticmethod
     def optimize_singleton_imputation_prior_strength(
@@ -385,7 +403,8 @@ class SingletonImputationOptimizer:
         negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
         negative_control_guide_control_population_total_normalized_counts_reps: List[float], 
         replicate_indices: List[int],
-        KL_guide_set_weights: Union[List[float], None]) -> Tuple[List[float], List[float]]:
+        KL_guide_set_weights: Union[List[float], None],
+        cores: int) -> Tuple[List[float], List[float]]:
 
         def weighted_objective_function_of_all_guides(rep_i: int,
                                                         experiment_guide_sets: ExperimentGuideSets,
@@ -393,7 +412,7 @@ class SingletonImputationOptimizer:
                                                         negative_control_guide_control_population_total_normalized_counts_reps: List[float],
                                                         KL_guide_set_weights: Union[List[float], None],
                                                         params: float):
-            KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = SingletonImputationOptimizer.objective_function_of_all_guides(rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
+            KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = SingletonImputationOptimizer.singleton_imputation_objective_function_of_all_guides(rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, params)
 
             if KL_guide_set_weights is None:
                 return KL_guide_imputation_score_total_combined_avg
@@ -423,7 +442,10 @@ class SingletonImputationOptimizer:
             optimizer_bounds_a = max_posterior_bounds_a / negative_control_guide_sample_population_total_normalized_counts_reps[rep_i]
             optimizer_bounds_b = max_posterior_bounds_b / negative_control_guide_control_population_total_normalized_counts_reps[rep_i]
             optimizer_upper_bound = np.min([optimizer_bounds_a, optimizer_bounds_b])
-            res = scipy.optimize.differential_evolution(weighted_objective_function_of_all_guides_p, bounds=[(0.000001, optimizer_upper_bound)], callback=store_values, maxiter= 10000) 
+            tolerance = optimizer_upper_bound / 1000
+            print("Singleton optimization upper bound: {}".format(optimizer_upper_bound))
+            print("Tolerance: {}".format(tolerance))
+            res = scipy.optimize.differential_evolution(weighted_objective_function_of_all_guides_p, bounds=[(0.000001, optimizer_upper_bound)], callback=store_values, maxiter= 10000, tol=0.005, workers=cores) 
 
             plt.scatter(param_vals, loss_vals)
             plt.xlabel("Prior Strength")
@@ -448,84 +470,84 @@ class SingletonImputationOptimizer:
         singleton_imputation_prior_strength_selected = np.asarray(singleton_imputation_prior_strength_selected).transpose().flatten()
         return singleton_imputation_prior_strength_selected
 
+def neighborhood_imputation_objective_function_of_each_guide_set(rep_i: int,
+                                            guide_set: List[Guide],
+                                            neighborhood_optimization_guide_sample_size: int,
+                                            experiment_guide_sets: ExperimentGuideSets,
+                                            negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+                                            negative_control_guide_control_population_total_normalized_counts_reps: List[float],
+                                            neighborhood_bandwidth: float,
+                                            deviation_weights: List[float],
+                                            params: Tuple[float, float]):
+    neighborhood_imputation_prior_strength_test, neighborhood_imputation_likelihood_strength_test = params
+    
+    if len(guide_set) == 0:
+        return 0,0,0
+
+    KL_guide_imputation_score_total: float = 0
+
+    neighborhood_optimization_guide_sample_size_i = len(guide_set) if neighborhood_optimization_guide_sample_size > len(guide_set) else neighborhood_optimization_guide_sample_size
+    sampled_guide_set = random.sample(guide_set, neighborhood_optimization_guide_sample_size_i)
+    for each_guide in sampled_guide_set:
+        # Get the posterior
+        imputation_posterior_alpha, imputation_posterior_beta, each_guide_sample_population_spatial_contribution_reps, each_guide_control_population_spatial_contribution_reps = ModelInference.perform_neighboorhood_score_imputation(each_guide, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength_test, neighborhood_imputation_likelihood_strength_test, [rep_i], neighborhood_bandwidth)
+
+        imputation_posterior_alpha = imputation_posterior_alpha[0]
+        imputation_posterior_beta = imputation_posterior_beta[0]
+
+        true_alpha = each_guide.sample_population_normalized_count_reps[rep_i]
+        true_beta = each_guide.control_population_normalized_count_reps[rep_i]
+
+        # Calculate KL divergence between the posterior and the likelihood
+        KL_guide_imputation_score: float = StatisticalHelperMethods.KL_beta(true_alpha, true_beta, imputation_posterior_alpha, imputation_posterior_beta)
+
+        # Add weight towards guides that deviate from the negative control.
+        KL_negative_control_deviation = 0
+        if each_guide.position is not None:
+            KL_negative_control_deviation: float = StatisticalHelperMethods.KL_beta(negative_control_guide_sample_population_total_normalized_counts_reps[rep_i], negative_control_guide_control_population_total_normalized_counts_reps[rep_i], each_guide_sample_population_spatial_contribution_reps[0], each_guide_control_population_spatial_contribution_reps[0])
+
+            KL_guide_imputation_score = (1+(deviation_weights[rep_i]*KL_negative_control_deviation))*KL_guide_imputation_score
+
+        # Add score to the main placeholder to get the final sum
+        KL_guide_imputation_score_total = KL_guide_imputation_score_total + KL_guide_imputation_score 
+        
+    return KL_guide_imputation_score_total, neighborhood_optimization_guide_sample_size_i, len(guide_set)
+
+
+def neighborhood_imputation_objective_function_of_all_guides(rep_i: int,
+                                    neighborhood_optimization_guide_sample_size: int,
+                                    experiment_guide_sets: ExperimentGuideSets,
+                                    negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+                                    negative_control_guide_control_population_total_normalized_counts_reps: List[float],
+                                    neighborhood_bandwidth: float,
+                                    deviation_weights: List[float],
+                                    params: Tuple[float, float]):
+    
+    # Iterate through each guide to test prior with tested weight
+
+    KL_guide_imputation_score_total_negative, negative_sampled_n, negative_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.negative_control_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
+
+    KL_guide_imputation_score_total_positive, positive_sampled_n, positive_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.positive_control_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
+
+    KL_guide_imputation_score_total_observation, observation_sampled_n, observation_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.observation_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
+    
+    total_n = (negative_total_n+positive_total_n+observation_total_n)
+    
+    KL_guide_imputation_score_total_negative_avg = np.inf if negative_sampled_n == 0 else KL_guide_imputation_score_total_negative/negative_sampled_n
+    KL_guide_imputation_score_total_positive_avg = np.inf if positive_sampled_n == 0 else KL_guide_imputation_score_total_positive/positive_sampled_n
+    KL_guide_imputation_score_total_observation_avg = np.inf if observation_sampled_n == 0 else KL_guide_imputation_score_total_observation/observation_sampled_n
+    KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (
+            (0 if negative_sampled_n == 0 else (KL_guide_imputation_score_total_negative/negative_sampled_n)*negative_total_n) + 
+            (0 if positive_sampled_n == 0 else (KL_guide_imputation_score_total_positive/positive_sampled_n)*positive_total_n) + 
+            (0 if observation_sampled_n == 0 else (KL_guide_imputation_score_total_observation/observation_sampled_n)*observation_total_n)
+            ) / total_n # NOTE: Because the optimization is performed on mini-batches, need to consider the mini-batch size when calculating the combined average score. 
+
+    return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
 
 class NeighborhoodImputationOptimizer:
-    @staticmethod
-    def objective_function_of_each_guide_set(rep_i: int,
-                                                guide_set: List[Guide],
-                                                neighborhood_optimization_guide_sample_size: int,
-                                                experiment_guide_sets: ExperimentGuideSets,
-                                                negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
-                                                negative_control_guide_control_population_total_normalized_counts_reps: List[float],
-                                                neighborhood_bandwidth: float,
-                                                deviation_weights: List[float],
-                                                params: Tuple[float, float]):
-        neighborhood_imputation_prior_strength_test, neighborhood_imputation_likelihood_strength_test = params
-        
-        if len(guide_set) == 0:
-            return 0,0,0
+    neighborhood_imputation_objective_function_of_each_guide_set = staticmethod(neighborhood_imputation_objective_function_of_each_guide_set)
+    neighborhood_imputation_objective_function_of_all_guides = staticmethod(neighborhood_imputation_objective_function_of_all_guides)
 
-        KL_guide_imputation_score_total: float = 0
-
-        neighborhood_optimization_guide_sample_size_i = len(guide_set) if neighborhood_optimization_guide_sample_size > len(guide_set) else neighborhood_optimization_guide_sample_size
-        sampled_guide_set = random.sample(guide_set, neighborhood_optimization_guide_sample_size_i)
-        for each_guide in sampled_guide_set:
-            # Get the posterior
-            imputation_posterior_alpha, imputation_posterior_beta, each_guide_sample_population_spatial_contribution_reps, each_guide_control_population_spatial_contribution_reps = ModelInference.perform_neighboorhood_score_imputation(each_guide, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength_test, neighborhood_imputation_likelihood_strength_test, [rep_i], neighborhood_bandwidth)
-
-            imputation_posterior_alpha = imputation_posterior_alpha[0]
-            imputation_posterior_beta = imputation_posterior_beta[0]
-
-            true_alpha = each_guide.sample_population_normalized_count_reps[rep_i]
-            true_beta = each_guide.control_population_normalized_count_reps[rep_i]
-
-            # Calculate KL divergence between the posterior and the likelihood
-            KL_guide_imputation_score: float = StatisticalHelperMethods.KL_beta(true_alpha, true_beta, imputation_posterior_alpha, imputation_posterior_beta)
-
-            # Add weight towards guides that deviate from the negative control.
-            KL_negative_control_deviation = 0
-            if each_guide.position is not None:
-                KL_negative_control_deviation: float = StatisticalHelperMethods.KL_beta(negative_control_guide_sample_population_total_normalized_counts_reps[rep_i], negative_control_guide_control_population_total_normalized_counts_reps[rep_i], each_guide_sample_population_spatial_contribution_reps[0], each_guide_control_population_spatial_contribution_reps[0])
-
-                KL_guide_imputation_score = (1+(deviation_weights[rep_i]*KL_negative_control_deviation))*KL_guide_imputation_score
-
-            # Add score to the main placeholder to get the final sum
-            KL_guide_imputation_score_total = KL_guide_imputation_score_total + KL_guide_imputation_score 
-            
-        return KL_guide_imputation_score_total, neighborhood_optimization_guide_sample_size_i, len(guide_set)
-
-
-    @staticmethod
-    def objective_function_of_all_guides(rep_i: int,
-                                        neighborhood_optimization_guide_sample_size: int,
-                                        experiment_guide_sets: ExperimentGuideSets,
-                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
-                                        negative_control_guide_control_population_total_normalized_counts_reps: List[float],
-                                        neighborhood_bandwidth: float,
-                                        deviation_weights: List[float],
-                                        params: Tuple[float, float]):
-        
-        # Iterate through each guide to test prior with tested weight
-
-        KL_guide_imputation_score_total_negative, negative_sampled_n, negative_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.negative_control_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
-
-        KL_guide_imputation_score_total_positive, positive_sampled_n, positive_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.positive_control_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
-
-        KL_guide_imputation_score_total_observation, observation_sampled_n, observation_total_n = NeighborhoodImputationOptimizer.objective_function_of_each_guide_set(rep_i, experiment_guide_sets.observation_guides, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
-        
-        total_n = (negative_total_n+positive_total_n+observation_total_n)
-        
-        KL_guide_imputation_score_total_negative_avg = np.inf if negative_sampled_n == 0 else KL_guide_imputation_score_total_negative/negative_sampled_n
-        KL_guide_imputation_score_total_positive_avg = np.inf if positive_sampled_n == 0 else KL_guide_imputation_score_total_positive/positive_sampled_n
-        KL_guide_imputation_score_total_observation_avg = np.inf if observation_sampled_n == 0 else KL_guide_imputation_score_total_observation/observation_sampled_n
-        KL_guide_imputation_score_total_combined_avg = np.inf if total_n == 0 else (
-                (0 if negative_sampled_n == 0 else (KL_guide_imputation_score_total_negative/negative_sampled_n)*negative_total_n) + 
-                (0 if positive_sampled_n == 0 else (KL_guide_imputation_score_total_positive/positive_sampled_n)*positive_total_n) + 
-                (0 if observation_sampled_n == 0 else (KL_guide_imputation_score_total_observation/observation_sampled_n)*observation_total_n)
-                ) / total_n # NOTE: Because the optimization is performed on mini-batches, need to consider the mini-batch size when calculating the combined average score. 
-
-        return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
-        
     @staticmethod
     def optimize_neighborhood_imputation_prior_strength(
         experiment_guide_sets: ExperimentGuideSets, 
@@ -535,6 +557,7 @@ class NeighborhoodImputationOptimizer:
         neighborhood_bandwidth: float, 
         deviation_weights: List[float], 
         KL_guide_set_weights: List[float],
+        cores: int,
         neighborhood_optimization_guide_sample_size: int = 50) -> Tuple[List[float], List[float]]:
 
 
@@ -546,7 +569,7 @@ class NeighborhoodImputationOptimizer:
                                         deviation_weights: List[float],
                                         guide_set_weights: Union[List[float], None], 
                                         params: Tuple[float, float]):
-            KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = NeighborhoodImputationOptimizer.objective_function_of_all_guides(rep_i, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
+            KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = NeighborhoodImputationOptimizer.neighborhood_imputation_objective_function_of_all_guides(rep_i, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
             
             if KL_guide_set_weights is None:
                 return KL_guide_imputation_score_total_combined_avg
@@ -570,7 +593,14 @@ class NeighborhoodImputationOptimizer:
                 loss_vals.append(f)
             
             # TODO: Set bounds as just positive - ask chatgpt how...
-            res = scipy.optimize.differential_evolution(weighted_objective_function_of_all_guides_p, bounds=[(0.000001, 5),(0.000001, 5)], callback=store_values, tol = 0.6) 
+            max_posterior_bounds_a = 100
+            max_posterior_bounds_b = 100
+            optimizer_bounds_a = max_posterior_bounds_a / negative_control_guide_sample_population_total_normalized_counts_reps[rep_i]
+            optimizer_bounds_b = max_posterior_bounds_b / negative_control_guide_control_population_total_normalized_counts_reps[rep_i]
+            optimizer_upper_bound = np.min([optimizer_bounds_a, optimizer_bounds_b])
+            tolerance = optimizer_upper_bound / 1000
+            print("Neighborhood optimization upper bound: {}".format(optimizer_upper_bound))
+            res = scipy.optimize.differential_evolution(weighted_objective_function_of_all_guides_p, bounds=[(0.000001, optimizer_upper_bound),(0.000001, optimizer_upper_bound)], callback=store_values, tol=0.005, workers=cores) 
 
             X=[param[0] for param in param_vals]
             Y=[param[1] for param in param_vals]
@@ -614,6 +644,7 @@ class ShrinkageOptimizer:
     def weighted_objective_function_of_all_guides(rep_i: int, 
                                             spatial_guides_for_fit: List[Guide],
                                             singleton_guides_for_fit: List[Guide], 
+                                            neighborhood_experiment_guide_sets: ExperimentGuideSets,
                                             negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
                                             negative_control_guide_control_population_total_normalized_counts_reps: List[float],
                                             neighborhood_imputation_prior_strength: List[float],
@@ -643,7 +674,7 @@ class ShrinkageOptimizer:
                 unweighted_prior_beta: float = np.asarray([negative_control_guide_control_population_total_normalized_counts_reps[rep_i]])
 
                 # If able to use spatial information, replace the unweighted priors with the spatial imputational posterior
-                imputation_posterior_alpha, imputation_posterior_beta, _, _ = ModelInference.perform_neighboorhood_score_imputation(each_guide, spatial_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength, neighborhood_imputation_likelihood_strength, [rep_i], neighborhood_bandwidth)
+                imputation_posterior_alpha, imputation_posterior_beta, _, _ = ModelInference.perform_neighboorhood_score_imputation(each_guide, neighborhood_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength, neighborhood_imputation_likelihood_strength, [rep_i], neighborhood_bandwidth)
 
                 # Propogate the imputation posterior to the shrinkage prior
                 unweighted_prior_alpha = imputation_posterior_alpha
@@ -741,6 +772,7 @@ class ShrinkageOptimizer:
             weighted_objective_function_of_all_guides_p = functools.partial(ShrinkageOptimizer.weighted_objective_function_of_all_guides, rep_i,
                                             spatial_guides_for_fit,
                                             singleton_guides_for_fit,
+                                            neighborhood_experiment_guide_sets,
                                             negative_control_guide_sample_population_total_normalized_counts_reps,
                                             negative_control_guide_control_population_total_normalized_counts_reps,
                                             neighborhood_imputation_prior_strength,
@@ -759,6 +791,7 @@ class ShrinkageOptimizer:
                 param_vals.append(x)
                 loss_vals.append(f)
 
+
             # TODO: Set bounds as just positive - ask chatgpt how...
             res = scipy.optimize.minimize(weighted_objective_function_of_all_guides_p, [20], bounds=[(0.000001, 1000)], method="TNC", callback=store_values) 
             
@@ -799,7 +832,8 @@ def perform_adjustment(
     shrinkage_prior_strength: Union[List[float], None] = None,
     monte_carlo_trials: int = 1000,
     posterior_estimator: str = "mean",
-    random_seed: Union[int, None] = None
+    random_seed: Union[int, None] = None,
+    cores=1
     ):
     raw_negative_control_guides = copy.deepcopy(negative_control_guides)
     raw_positive_control_guides = copy.deepcopy(positive_control_guides)
@@ -896,7 +930,7 @@ def perform_adjustment(
     if enable_neighborhood_prior:
         if (neighborhood_imputation_prior_strength is None) or (neighborhood_imputation_likelihood_strength is None):
             print("Optimizing neighborhood imputation weights")
-            neighborhood_imputation_prior_strength,neighborhood_imputation_likelihood_strength = NeighborhoodImputationOptimizer.optimize_neighborhood_imputation_prior_strength(neighborhood_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, replicate_indices, neighborhood_bandwidth, deviation_weights, KL_guide_set_weights)
+            neighborhood_imputation_prior_strength,neighborhood_imputation_likelihood_strength = NeighborhoodImputationOptimizer.optimize_neighborhood_imputation_prior_strength(neighborhood_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, replicate_indices, neighborhood_bandwidth, deviation_weights, KL_guide_set_weights, cores)
             print("Selected neighborhood imputation weights: prior={}, likelihood={}".format(neighborhood_imputation_prior_strength, neighborhood_imputation_likelihood_strength))
         
     if singleton_imputation_prior_strength is None:
@@ -906,7 +940,8 @@ def perform_adjustment(
             negative_control_guide_sample_population_total_normalized_counts_reps, 
             negative_control_guide_control_population_total_normalized_counts_reps,
             replicate_indices,
-            KL_guide_set_weights)
+            KL_guide_set_weights,
+            cores)
         print("Selected singleton imputation weights: {}".format(singleton_imputation_prior_strength))
 
 
