@@ -2,7 +2,7 @@
 from typing import List, Union, Tuple
 from matplotlib import pyplot as plt
 import numpy as np
-from CrisprShrinkage import CrisprShrinkageResult, Guide
+from .CrisprShrinkage import CrisprShrinkageResult, Guide
 import matplotlib.pyplot as plt
 from scipy import stats
 from dataclasses import dataclass
@@ -50,7 +50,9 @@ class CrisprShrinkageVisualizationInput:
 def prepare_crispr_shrinkage_visualization_input(crispr_shrinkage_result: CrisprShrinkageResult):
     replicate_indices = range(crispr_shrinkage_result.num_replicates)
     explanatory_guides = [guide for guide in np.concatenate([crispr_shrinkage_result.adjusted_observation_guides, crispr_shrinkage_result.adjusted_positive_control_guides, crispr_shrinkage_result.adjusted_negative_control_guides]) if guide.is_explanatory is True]
-
+    
+    for guide in explanatory_guides:
+        assert guide.position is not None, "All explanatory guides (.is_explanatory==True) must contain a position"
 
     observational_position = np.asarray([guide.position for guide in crispr_shrinkage_result.adjusted_observation_guides])
     positive_positions = np.asarray([guide.position for guide in crispr_shrinkage_result.adjusted_positive_control_guides])
@@ -132,6 +134,9 @@ def prepare_crispr_shrinkage_visualization_input(crispr_shrinkage_result: Crispr
 
     return crispr_shrinkage_visualization_input
 
+def normalize_shape_size(input_shape_sizes: List[Union[int,float]], max_shape_size: Union[int,float]):
+    input_shape_sizes = np.asarray(input_shape_sizes)
+    return ((input_shape_sizes - input_shape_sizes.min()) / (input_shape_sizes.max() - input_shape_sizes.min())) * max_shape_size
 
 def visualize_lfc_histogram(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
     plt.hist(crispr_shrinkage_visualization_input.negative_control_lfc, density=True, label="Negative")
@@ -145,7 +150,7 @@ def visualize_lfc_by_count(crispr_shrinkage_visualization_input: CrisprShrinkage
     for rep_i in crispr_shrinkage_visualization_input.replicate_indices:
         plt.scatter(crispr_shrinkage_visualization_input.observational_count_rep[rep_i], crispr_shrinkage_visualization_input.observational_raw_lfc_rep[rep_i], alpha=0.6, label="observation")
         plt.scatter(crispr_shrinkage_visualization_input.positive_control_count_rep[rep_i], crispr_shrinkage_visualization_input.positive_control_raw_lfc_rep[rep_i], alpha=0.6, label="positive")
-        plt.scatter(crispr_shrinkage_visualization_input.negative_control_count_rep[rep_i], crispr_shrinkage_visualization_input.negative_raw_lfc_rep[rep_i], alpha=0.6, label="negative")
+        plt.scatter(crispr_shrinkage_visualization_input.negative_control_count_rep[rep_i], crispr_shrinkage_visualization_input.negative_control_raw_lfc_rep[rep_i], alpha=0.6, label="negative")
         plt.xlabel("Total Normalized Count")
         plt.ylabel("Raw LFC")
         plt.title("Replicate {}".format(rep_i+1))
@@ -173,23 +178,31 @@ def visualize_raw_vs_adjusted_score_scatter(crispr_shrinkage_visualization_input
         plt.title("Replicate {}".format(rep_i + 1))
         plt.show()
 
-def visualize_raw_and_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+def visualize_raw_and_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput, max_shape_size = 10):
     for rep_i in crispr_shrinkage_visualization_input.replicate_indices:
-        plt.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i])
+        plt.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size))
         plt.title("Adjusted LFC")
         plt.show()
-        plt.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_raw_lfc_rep[rep_i], s=crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i])
+        plt.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_raw_lfc_rep[rep_i], s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size))
         plt.title("Raw LFC")
         plt.show()
 
-def visualize_combined_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+def visualize_combined_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput, max_shape_size = 10):
+    explanatory_positions = np.asarray(crispr_shrinkage_visualization_input.explanatory_positions)
+    explanatory_lfc = np.asarray(crispr_shrinkage_visualization_input.explanatory_lfc)
+
+    
+    explanatory_positions_argsort = np.argsort(explanatory_positions)
+    explanatory_positions = explanatory_positions[explanatory_positions_argsort]
+    explanatory_lfc = explanatory_lfc[explanatory_positions_argsort]
+
     fig = plt.figure(figsize=(20, 4))
     ax = fig.add_subplot(111)
 
     for rep_i in crispr_shrinkage_visualization_input.replicate_indices:
-        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], alpha=0.3, label="Replicate {}".format(rep_i))
-    ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc, marker="s", s=2, label="Combined")
-    ax.plot(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc,color="red", alpha=0.2)
+        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size), alpha=0.3, label="Replicate {}".format(rep_i))
+    ax.scatter(explanatory_positions, explanatory_lfc, marker="s", s=2, label="Combined")
+    ax.plot(explanatory_positions, explanatory_lfc, color="red", alpha=0.2)
     ax.set_title("Region Scores")
     ax.set_xlabel("Coordinate")
     ax.set_ylabel("Adjusted LFC")
@@ -197,31 +210,43 @@ def visualize_combined_adjusted_score_by_position_scatter(crispr_shrinkage_visua
     plt.show()
 
 def visualize_combined_score_credible_interval_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+    explanatory_positions = np.asarray(crispr_shrinkage_visualization_input.explanatory_positions)
+    explanatory_lfc = np.asarray(crispr_shrinkage_visualization_input.explanatory_lfc)
+
+    
+    explanatory_positions_argsort = np.argsort(explanatory_positions)
+    explanatory_positions = explanatory_positions[explanatory_positions_argsort]
+    explanatory_lfc = explanatory_lfc[explanatory_positions_argsort]
+
+
     explanatory_lfc_CI_low = np.asarray([CI[0] for CI in crispr_shrinkage_visualization_input.explanatory_lfc_CI])
     explanatory_lfc_CI_up = np.asarray([CI[1] for CI in crispr_shrinkage_visualization_input.explanatory_lfc_CI])
     
+    explanatory_lfc_CI_low = explanatory_lfc_CI_low[explanatory_positions_argsort]
+    explanatory_lfc_CI_up = explanatory_lfc_CI_up[explanatory_positions_argsort]
+
     fig = plt.figure(figsize=(20, 4))
     ax = fig.add_subplot(111)
 
-    ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc, marker="s",color="black", s=2, label="Combined")
+    ax.scatter(explanatory_positions, explanatory_lfc, marker="s",color="black", s=2, label="Combined")
 
-    ax.errorbar(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc, (crispr_shrinkage_visualization_input.explanatory_lfc-explanatory_lfc_CI_low, explanatory_lfc_CI_up-crispr_shrinkage_visualization_input.explanatory_lfc), solid_capstyle='projecting', capsize=1, alpha=0.3)
-    ax.plot(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc,color="red", alpha=0.2)
+    ax.errorbar(explanatory_positions, explanatory_lfc, (explanatory_lfc-explanatory_lfc_CI_low, explanatory_lfc_CI_up-explanatory_lfc), solid_capstyle='projecting', capsize=1, alpha=0.3)
+    ax.plot(explanatory_positions, explanatory_lfc, color="red", alpha=0.2)
     ax.set_title("Region Scores")
     ax.set_xlabel("Coordinate")
     ax.set_ylabel("Adjusted LFC")
     ax.legend()
     plt.show()
 
-def visualize_all_adjusted_score_credible_interval_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+def visualize_all_adjusted_score_credible_interval_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput, max_shape_size = 10):
     for rep_i in crispr_shrinkage_visualization_input.replicate_indices:
         fig = plt.figure(figsize=(20, 4))
         ax = fig.add_subplot(111)
         explanatory_lfc_rep_rep_standardized = stats.zscore(crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i])
         explanatory_raw_lfc_rep_standardized = stats.zscore(crispr_shrinkage_visualization_input.explanatory_raw_lfc_rep[rep_i])
 
-        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions,explanatory_lfc_rep_rep_standardized, s=np.asarray(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i])*0.1, color="blue", alpha=0.85, label="Adjusted")
-        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions,explanatory_raw_lfc_rep_standardized, s=np.asarray(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i])*0.1, color="black", alpha=0.85, label="Raw") # Can color based on which of the guides are in the observational set and which are not (specifically which are in the positive, negative.
+        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions,explanatory_lfc_rep_rep_standardized, s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size), color="blue", alpha=0.85, label="Adjusted")
+        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions,explanatory_raw_lfc_rep_standardized, s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size), color="black", alpha=0.85, label="Raw") # Can color based on which of the guides are in the observational set and which are not (specifically which are in the positive, negative.
         for guide_i in range(len(crispr_shrinkage_visualization_input.explanatory_positions)):
             ax.arrow(crispr_shrinkage_visualization_input.explanatory_positions[guide_i], explanatory_raw_lfc_rep_standardized[guide_i],0, explanatory_lfc_rep_rep_standardized[guide_i] - explanatory_raw_lfc_rep_standardized[guide_i], head_width=1, head_length=0.2, alpha=0.4, length_includes_head=True, color="black")
 
@@ -231,7 +256,7 @@ def visualize_all_adjusted_score_credible_interval_by_position_scatter(crispr_sh
         ax.set_title("Replicate {}".format(rep_i+1))
         plt.show()
 
-def visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+def visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput, max_shape_size=10):
     explanatory_lfc_CI_low = np.asarray([CI[0] for CI in crispr_shrinkage_visualization_input.explanatory_lfc_CI])
     explanatory_lfc_CI_up = np.asarray([CI[1] for CI in crispr_shrinkage_visualization_input.explanatory_lfc_CI])
     
@@ -240,7 +265,7 @@ def visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_v
     ax = fig.add_subplot(111)
 
     for rep_i in crispr_shrinkage_visualization_input.replicate_indices:
-        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=np.asarray(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i])*0.5, alpha=0.3, label="Replicate {}".format(rep_i))
+        ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc_rep[rep_i], s=normalize_shape_size(crispr_shrinkage_visualization_input.explanatory_count_rep[rep_i], max_shape_size), alpha=0.3, label="Replicate {}".format(rep_i))
     ax.scatter(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc, marker="s",color="black", s=10, label="Combined")
     ax.errorbar(crispr_shrinkage_visualization_input.explanatory_positions, crispr_shrinkage_visualization_input.explanatory_lfc, (crispr_shrinkage_visualization_input.explanatory_lfc-explanatory_lfc_CI_low, explanatory_lfc_CI_up-crispr_shrinkage_visualization_input.explanatory_lfc), solid_capstyle='projecting', capsize=1, alpha=0.3, linestyle='')
 
@@ -250,14 +275,15 @@ def visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_v
     ax.legend()
     plt.show()
 
-def visualize_all(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
+def visualize_all(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput, max_shape_size=10):
     visualize_lfc_histogram(crispr_shrinkage_visualization_input)
+    visualize_lfc_by_count(crispr_shrinkage_visualization_input)
     visualize_raw_vs_adjusted_score_scatter(crispr_shrinkage_visualization_input)
-    visualize_raw_and_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input)
-    visualize_combined_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input)
+    visualize_raw_and_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input, max_shape_size)
+    visualize_combined_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input, max_shape_size)
     visualize_combined_score_credible_interval_scatter(crispr_shrinkage_visualization_input)
-    visualize_all_adjusted_score_credible_interval_by_position_scatter(crispr_shrinkage_visualization_input)
-    visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input)
+    visualize_all_adjusted_score_credible_interval_by_position_scatter(crispr_shrinkage_visualization_input, max_shape_size)
+    visualize_standardized_adjusted_score_by_position_scatter(crispr_shrinkage_visualization_input, max_shape_size)
 
 def prepare_reporting_df(crispr_shrinkage_visualization_input: CrisprShrinkageVisualizationInput):
     pass
\ No newline at end of file
